package LFastaIO;

# Time-stamp: <2005-02-17 17:30:28 kasper>

use strict;
use warnings;
use LFasta;
use Carp;

require Exporter;
# use AutoLoader qw(AUTOLOAD);

our @ISA = qw(Exporter);

# Items to export into callers namespace by default. Note: do not export
# names by default without a very good reason. Use EXPORT_OK instead.
# Do not simply export all your public functions/methods/constants.

# This allows declaration	use LFastaIO ':all';
# If you do not need this, moving things directly into @EXPORT or @EXPORT_OK
# will save memory.
our %EXPORT_TAGS = ( 'all' => [ qw(

) ] );

our @EXPORT_OK = ( @{ $EXPORT_TAGS{'all'} } );

our @EXPORT = qw(

);

our $VERSION = '0.01';


sub new {
  my $invocant = shift;
  my $class = ref($invocant) || $invocant;
  my $self = {};
  my $tmp = {@_};

  # In case I accidentally supply attributes in the option style
  # bioperl uses:
  while (my ($key, $value) = each %$tmp) {
    $key =~ s/^-*(.+)/$1/;
    $$self{$key} = $value;
  }

  $$self{fh} && $$self{file} &&
    croak "Don't specify both file and filehandle";

  # Make a file handle or get the one specified:
  my $handle;
  if ($$self{file}) {
    open $handle, "$$self{file}" or croak $!;
  } elsif ($$self{fh}) {
    $handle = $$self{fh};
  } else {
    croak "You have to specify either file or filehandle";
  }

  # Format defualts to LFasta:
  $$self{outformat} ||= 'LFasta';

  # Default is no plp printing:
  $$self{plpprint} ||= 0;

  # Create a Bio::SeqIO file handle if format is not LFasta:
  if ($$self{outformat} ne 'LFasta') {
    use Bio::SeqIO;
    open $$self{fh}, $$self{file} if $$self{file};
    $$self{seqio} = Bio::SeqIO->newFh(-fh => $$self{fh},
                                      -format => lc($$self{outformat}));
  }
  $$self{seqio} ||= '';

  # Do the tie and bless:
  local *FH;
  my $fh = *FH;
  tie($fh, 'LFastaIO', $handle, $$self{seqio}, $$self{outformat}, $$self{plpprint});

  bless \$fh, $class;
}

# sub outformat {
#   my $self = shift;
#   if (@_) {
#     $$self{outformat} = shift;
#     $$self{seqio} = Bio::SeqIO->newFh(-fh => $$self{fh},
#                                       -format => lc($$self{outformat}));
#   }
#   return $$self{outformat};
# }

# # This declaration guarantees that we get our own autoloader:
# use subs qw(file);
# 
# sub AUTOLOAD {
#   my $self = shift;
#   # Only handle indtance methods, not class methods.
#   croak "$self not an object" unless ref($self);
#   my $name = our $AUTOLOAD;
#   $name =~ s/^.*:([^:]+)$/$1/;
#   return if $name =~ /::DESTROY$/;
#   unless (exists $self->{$name}) {
#     croak "Can't access '$name' field in $self";
#   }
#   if (@_) {
#     return $self->{$name} = shift;
#   }
#   else {
#     return $self->{$name};
#   }
# }

sub _savedheader {
  my $self = shift;
  if ($$self{saved_header}) {
    my $h = $$self{saved_header};
    $$self{saved_header} = undef;
    return $h;
  } else {
    return undef;
  }
}
sub _getentry {
  my $self = shift;

  my $fh = $$self{handle};

  my @entry = ();
  my %labels = ();
  my @plpdata = ();

  my $nl = 3;

  # Keep track of whether we are reading into a new entry:
  my $got_a_header = 0;

  my $id;

  while (my $line = $$self{saved_header} ? _savedheader($self) : <$fh>) {
    chomp $line;

    # If we get a header line, we want it to be the first one we see:
    if ($line =~ /^>/ && not $got_a_header) {
      $got_a_header = 1;
      $line =~ />(\S+)/ or die;
      $id = $1;
      $entry[0] = $line;
    } elsif ($line =~ /^>/) {
      # Seems wee are getting ahead of ourself..
      $$self{saved_header} = $line;
      last;
    } elsif ($got_a_header) {
      if ($line =~ /^%/) {
        # Append comments to the id line

      } elsif ($id && $line =~ /^# $id/) {
        # First plp line:
        push @plpdata, $line;
        next;
      } elsif ($line =~ /^# {9}\S\s/) {
        # Second plp line:
        push @plpdata, $line;
        next;
      } elsif ($line =~ /^\S\s\S\s+(?:\s+(?:(?i)(?:[+-]?)(?:(?=[0123456789]|[.])(?:[0123456789]*)(?:(?:[.])(?:[0123456789]{0,}))?)(?:(?:[Ee])(?:(?:[+-]?)(?:[0123456789]+))|))){2,}/) {
        # Other plp lines:
        push @plpdata, $line;
        next;
      } else {
        # Determine which other entry in the array to write to:
        my $k;
        if ($line =~ /^#/) {
          if ( !defined($entry[2]) ) {
            $entry[2] = "";
          }
          $line =~ s/^#//;
          $k = 2;
        } elsif ($line =~ /^\?/) {
          my $label = substr($line,1,1);
          if (defined($labels{$label})) {
            $k = $labels{$label};
          } else {
            $entry[$nl] = $label;
            $labels{$label} = $nl + 1;
            $k = $nl + 1;
            $nl += 2;
            $entry[$k] = "";
          }
          $line =~ s/^\?.//;
        } else {
          $k = 1;
        }
        $line =~ s/ //g;
        $entry[$k] .= $line;
      }
    }
    # Save new possition of file handle:
  }

  return undef unless $entry[0] && $entry[1];

  my $lfa = LFasta->new( header => $entry[0],
                         seq => $entry[1],
                         labels => $entry[2],
                         plpdata => \@plpdata );

  for (my $i = 4; defined $entry[$i]; $i += 2) {
    my $tag = '';
    if (defined $entry[$i-1]) {
      $tag = $entry[$i-1];
    }
    $lfa->add_pred( prefix => $tag,
                    pred => $entry[$i] );
  }

  return $lfa;
}

sub _printLFasta {
  my $self = shift;
  my $lfa = shift;

  my $linelen = 70;
  my $entrystr = '';
   $entrystr .= ">" if ( $lfa->header !~ /^>/ );
   $entrystr .= $lfa->header . "\n";
  my $indentation = "";
  my $labelprefix;
  if (defined($lfa->labels)) {
    $indentation = '  ';
    $labelprefix = '# '
  }
  if ($lfa->labels) {
    $indentation = '   ';
    $labelprefix = '#  ';
  }
  my $length = length($lfa->seq);
  for (my $i=0; $i<$length; $i += $linelen) {
    $entrystr .= $indentation . substr($lfa->seq,$i,$linelen) . "\n";
    if (defined($lfa->labels)) {
      $entrystr .= $labelprefix . substr($lfa->labels,$i,$linelen) . "\n";
    }
    my $predictions = $lfa->preds;
    for my $pred (@$predictions) {
      if (defined $$pred[0]) {
        $entrystr .= "?$$pred[0] ";
      } else {
        $entrystr .= '?  ';
      }
      $entrystr .=  substr($$pred[1],$i,$linelen) . "\n";
    }
    $entrystr .= "\n" if @$predictions;
  }

  if ($$self{plpprint}) {
    my $plpdata = $lfa->plpdata();
    $entrystr .= "\n";
    for my $line (@$plpdata) {
      $entrystr .= "$line\n"
    }
  }

  return $entrystr;
}

# Returns a bioperl Seq obj:
sub _LFasta2Seq {
  my $lfa = shift;

  my $seq = Bio::Seq->new(-id => $lfa->id,
                          -desc => $lfa->desc,
                          -seq => $lfa->seq);

  # Get the gff lines:
  my $gfflines = $lfa->get_gff_features;
  for my $gffstr (@$gfflines) {
    # Turn them into features:
    my $feat = Bio::SeqFeature::Generic->new(-gff_string => $gffstr);
    $seq->add_SeqFeature($feat);
  }
  return $seq;
}

sub _LFasta2FastaSeq {
  my $lfa = shift;

  my $seq = Bio::Seq->new(-id => $lfa->id,
                          -desc => $lfa->desc,
                          -seq => $lfa->seq);
  return $seq;
}

# TIEHANDLE classname, LIST
# This is the constructor for the class.  That means it is expected to
# return a blessed reference of some sort. The reference can be used
# to hold some internal information.
sub TIEHANDLE {
  my $class = shift;
  my $handle = shift;
  my $seqio = shift;
  my $outformat = shift;
  my $plpprint = shift;
  bless {handle => $handle,
         seqio => $seqio,
         outformat => $outformat,
         plpprint => $plpprint}, $class;
}

# PRINT this, LIST
# This method will be triggered every time the tied handle is printed
# to with the "print()" function.  Beyond its self reference it also
# expects the list that was passed to the print function.
sub PRINT {
  my $self = shift;
  my $handle = $$self{handle};
  my $seqio = $$self{seqio};
  for my $lfa (@_) {
    ref($lfa) eq 'LFasta'
      or croak "Don't use an LFastaIO to print anything but LFasta objects";
    if ($$self{outformat} eq 'LFasta') {
      my $str = _printLFasta($self, $lfa);
      print $handle $str;
#      print $handle $lfa->printLFasta;
    } elsif ($$self{outformat} eq 'Fasta') {
      my $seq = _LFasta2FastaSeq($lfa);
      print $seqio $seq;
    } else {
      my $seq = _LFasta2Seq($lfa);
      print $seqio $seq;
    }
  }
}

# READLINE this
# This method will be called when the handle is read from via <HAN-
# DLE>.  The method should return undef when there is no more data.
sub READLINE {
  my $self = shift;
  my $lfa = _getentry($self);
  return $lfa
}

# CLOSE this
# This method will be called when the handle is closed via the
# "close" function.
sub CLOSE { print "CLOSE called.\n" }

# UNTIE this
# As with the other types of ties, this method will be called when
# "untie" happens.  It may be appropriate to "auto CLOSE" when this
# occurs.  See "The "untie" Gotcha" below.

# # DESTROY this
# # As with the other types of ties, this method will be called when
# # the tied handle is about to be destroyed. This is useful for debug-
# # ging and possibly cleaning up.
# sub DESTROY {
#   my $self = shift;
#   my $fh = $$self{handle};
#   close $fh;
# }

1;

__END__

=head1 NAME

LFastaIO - An IO interface for LFasta.

=head1 SYNOPSIS

 use LFastaIO;

 # Create a LFasta obj. from scratch:
 my $lfa = LFasta->new(header => 'ID And then some description',
                       seq => 'AAAGGTGTCCACCGGCGTT',
                       labels => 'xxxxHHHHHHHHxxxxxxx';

 # or

 my $lfa = LFasta->new(id => 'ID',
                       desc => 'Some description',
                       seq => 'AAAGGTGTCCACCGGCGTT',
                       labels => xxxxHHHHHHHHxxxxxxx;

 # Note that the sequence and the label string must have equal length.

 # Read from an EMBL file:
 my $in = LFastaIO->new(file => "inputfile.lfa");

 # Output to STDOUT in LFasta format:
 my $out = LFastaIO->new(fh => \*STDOUT, outformat => 'EMBL');

 # Loop over and reformat the sequence entries SeqIO style:
 print $out while <$in>;


=head1 DESCRIPTION

LFastaIO is to LFasta what SeqIO is to Seq. It works in much the same
way, but does only support the filehandel emulation for input and
output. So LFastaIO->new corresponds to Bio::SeqIO->newFh. As of now
the module only supports LFasta as input format. For output formats
other than LFasta, it uses the facilities of SeqIO.

=head1 METHODS

The object methods.

=over 2

=item new

 Title   : new
 Usage   : $fh = LFastaIO->newFh(file => 'filename')
 Function: Creates a tied file handle.
 Example : $in = LFastaIO->new(file => 'infile',);
           $out = LFastaIO->new(file => '>outfile',
                                outformat => 'format');
           $lfa = <$in>;    # read a LFasta object
           print $out $lfa; # write a LFasta object
 Returns : filehandle tied to the LFastaIO class.
 Args    :

=back

=head1 DEPENDENCIES

The Bioperl distribution. Specifically Bio::SeqIO.

=head1 SEE ALSO

This module is a stand alone extention to Bioperl, and does not work
nor make sense without it.

The LFasta module.

=head1 AUTHOR

Kasper Munch / Anders Krogh

=head1 COPYRIGHT AND LICENSE

Copyright (C) 2004 by Kasper Munch

This library is free software; you can redistribute it and/or modify
it under the same terms as Perl itself, either Perl version 5.8.3 or,
at your option, any later version of Perl 5 you may have available.


=cut
