#!/usr/bin/perl

use warnings;
use strict;
use Test::More;

BEGIN { use_ok('LFastaIO') };

plan tests => 14;

# Test that the module can do the methods:
#can_ok('LFastaIO', qw/ new  _getentry _printLFasta _LFasta2Seq  TIEHANDLE PRINT READLINE CLOSE DESTROY/);
can_ok('LFastaIO', qw/ new  _getentry _printLFasta _LFasta2Seq TIEHANDLE PRINT READLINE CLOSE /);

# Check input file handle using a file handle:
open my $in, 't/1.lfa' or die $!;
my $lfa_in = LFastaIO->new(fh => $in);
ok( defined $lfa_in, 'new() returns defined');
ok( $lfa_in->isa('LFastaIO'), 'The class is right');
is( ref($lfa_in), 'LFastaIO', 'The type is right');

# Check output file handle using a file handle:
open my $out, '>t/tmp.lfa' or die $!;
my $lfa_out = LFastaIO->new(fh => $out);
ok( defined $lfa_out, 'new() returns defined');
ok( $lfa_out->isa('LFastaIO'), 'The class is right');
is( ref($lfa_out), 'LFastaIO', 'The type is right');

# Undef again:
$lfa_in = undef;
$lfa_out = undef;

# Try using files:
$lfa_in = LFastaIO->new(file => 't/1.lfa');
ok( defined $lfa_in, 'new() returns defined');
ok( $lfa_in->isa('LFastaIO'), 'The class is right');
is( ref($lfa_in), 'LFastaIO', 'The type is right');

$lfa_out = LFastaIO->new(file => '>t/tmp.lfa');
ok( defined $lfa_out, 'new() returns defined');
ok( $lfa_out->isa('LFastaIO'), 'The class is right');
is( ref($lfa_out), 'LFastaIO', 'The type is right');

# See if output is identical to input:
while (<$lfa_in>) {
  print $lfa_out $_;
}
ok(!`diff t/1.lfa t/tmp.lfa`, 'Input LFasta and output LFasta indentical')
  or diag('It seems that the output LFasta is not identical to the input LFasta.');

